import numpy as np
import pandas as pd
from .. import GenotypePhenotypeMap


def mask(gpm, mask_fraction):
    """Create a new GenotypePhenotypeMap object from a random subset of another
    GenotypePhenotypeMap.


    Returns
    -------
    true_mask_fraction : float
        the actual fraction used, since the space is discrete and likely
        won't be the exact fraction given.
    GenotypePhenotypeMap :
        the new genotype-phenotype map.
    """
    if mask_fraction > 1 or mask_fraction < 0:
        raise Exception("mask_fraction must between between 0 and 1.")

    # Calculate the number of genotypes to select
    number_to_choose = int((1 - mask_fraction) * gpm.n)

    # Calculate the true fraction (since this is a discrete space.)
    true_mask_fraction = 1 - float(number_to_choose) / gpm.n

    # Randomly choose genotypes
    index = np.random.choice(gpm.index, number_to_choose, replace=False)

    # Check n_replicates datatype
    if type(gpm.n_replicates) == int:
        n_replicates = gpm.n_replicates
    elif type(gpm.n_replicates) == pd.Series or
    type(gpm.n_replicates) == np.ndarray:
        n_replicates = gpm.n_replicates[index].reset_index(drop=True)
    else:
        raise Exception("n_replicates are not a valid dtype.")

    # return Subset genotype
    return true_mask_fraction, GenotypePhenotypeMap(
        gpm.wildtype,
        gpm.genotypes[index].reset_index(drop=True),
        gpm.phenotypes[index].reset_index(drop=True),
        mutations=gpm.mutations,
        stdeviations=gpm.stdeviations[index].reset_index(drop=True),
        n_replicates=n_replicates)



import numpy as np
import itertools as it

from gpmap.gpm import GenotypePhenotypeMap
from gpmap import utils
from .base import random_mutation_set, BaseSimulation


class NKSimulation(BaseSimulation):
    """Generate genotype-phenotype map from NK fitness model. Creates a table
    with binary sub-sequences that determine the order of epistasis in the
    model.

    The NK fitness landscape is created using a table with binary, length-K,
    sub-sequences mapped to random values. All genotypes are binary with
    length N. The fitness of a genotype is constructed by summing the values
    of all sub-sequences that make up the genotype using a sliding window
    across the full genotype.

    For example, imagine an NK simulation with N=5 and K=2. To construct
    the fitness for the 01011 genotype, select the following sub-sequences
    from an NK table "01", "10", "01", "11", "10". Sum their values together.

    Parameters
    ----------

    Attributes
    ----------
    nk_table : dict
        table with binary sub-sequences as keys which are used to construct
        phenotypes following an NK routine
    keys : array
        array of keys in NK table.
    values : array
        array of values in the NK table.
    """

    def __init__(self, wildtype, mutations, K, k_range=(0, 1),
                 *args, **kwargs):
        super(NKSimulation, self).__init__(wildtype, mutations,
                                           *args, **kwargs)
        # Set parameters
        self.set_order(K)
        self.set_random_values(k_range=k_range)
        self.build()

    @property
    def nk_table(self):
        """NK table mapping binary sequence to value."""
        return self.map("keys", "values")

    @property
    def keys(self):
        """NK table keys.
        """
        return self._keys

    @property
    def values(self):
        """NK table values
        """
        return self._values

    def set_order(self, K):
        """Set the order (K) of the NK model.
        """
        self.K = K
        # point to order
        self.order = self.K
        self._keys = np.array(["".join(r) for r in
                               it.product('01', repeat=self.K)])
        # Reset phenotypes
        self.data['phenotypes'] = np.empty(self.n, dtype=float)

    def set_random_values(self, k_range=(0, 1)):
        """Set the values of the NK table by drawing from a uniform
        distribution between the given k_range.
        """
        if hasattr(self, "keys") is False:
            raise Exception("Need to set K first. Try `set_order` method.")
        self._values = np.random.uniform(k_range[0], k_range[1],
                                         size=len(self.keys))
        self.build()

    def set_table_values(self, values):
        """Set the values of the NK table from a list/array of values.
        """
        if len(values) != len(self.keys):
            raise Exception("Length of the values do not equal the length of "
                            "NK keys. "
                            "Length of keys is : %d" % (len(self.keys),))
        self._values = values
        self.build()

    def build(self):
        """Build phenotypes from NK table.
        """
        nk_table = self.nk_table
        # Check for even interaction
        neighbor = int(self.order / 2)
        if self.order % 2 == 0:
            pre_neighbor = neighbor - 1
        else:
            pre_neighbor = neighbor
        # Use NK table to build phenotypes
        phenotypes = np.zeros(self.n, dtype=float)
        for i in range(len(self.genotypes)):
            f_total = 0
            for j in range(self.length):
                if j - pre_neighbor < 0:
                    pre = self.binary[i][-pre_neighbor:]
                    post = self.binary[i][j:neighbor + j + 1]
                    f = "".join(pre) + "".join(post)
                elif j + neighbor > self.length - 1:
                    pre = self.binary[i][j - pre_neighbor:j + 1]
                    post = self.binary[i][0:neighbor]
                    f = "".join(pre) + "".join(post)
                else:
                    f = "".join(
                        self.binary[i][j - pre_neighbor:j + neighbor + 1])
                f_total += nk_table[f]
            phenotypes[i] = f_total
        self.data.phenotypes = phenotypes



from .random import RandomPhenotypesSimulation
from .nk import NKSimulation
from .hoc import HouseOfCardsSimulation
from .fuji import MountFujiSimulation
from .multipeak_fuji import MultiPeakMountFujiSimulation



import numpy as np
import pandas as pd
from .gpm import GenotypePhenotypeMap

# ----------------------------------------------------------
# Sampling from Genotype-phenotype
# ----------------------------------------------------------

def split_data(gpm, fraction=1.0):
    """"""
    fraction = 0.6
    n = int(gpm.n * fraction)
    frac = n / gpm.n

    index = np.arange(0, gpm.n, dtype=int)
    np.random.shuffle(index)

    # Split data.
    train_set = gpm.data.loc[index[:n]]
    test_set = gpm.data.loc[~gpm.data.index.isin(train_set.index)]

def sample_genotypes(gpm, fraction=1.0, size=None):
    """Return new GenotypePhenotypeMap that's at genotype sampling of given
    GenotypePhenotypeMap. Samples without replacement.

    The fraction will round to the lowest whole number of genotypes.
    If size is given, fraction is ignored
    """
    # Shorten gpm.data
    d = gpm.data

    # Determine size
    n = len(d)
    if size is None:
        size = int(fraction * n)
    elif size > n:
        Exception('Size exceeds the number of genotypes.')

    # Sample genotypes
    index = np.random.choice(d.index, size=size, replace=False)

    # Return GenotypePhenotypeMap
    return GenotypePhenotypeMap(gpm.wildtype,
                                d.genotypes[index],
                                d.phenotypes[index],
                                mutations=gpm.mutations,
                                stdeviations=d.stdeviations[index],
                                n_replicates=d.n_replicates[index])


def sample_phenotypes(gpm, n_replicates=1):
    """Return new GenotypePhenotypeMap that's a phenotype sampling of given
    GenotypePhenotypeMap. Samples without replacement.

    Samples from normal distributions around each phenotype.
    """
    d = gpm.data

    # Draw samples
    samples = np.random.normal(loc=d.phenotypes,
                               scale=d.stdeviations,
                               size=(n_replicates, len(d.phenotypes)))

    phenotypes = np.mean(samples, axis=0)
    stdeviations = np.std(samples, axis=0)

    # Return GenotypePhenotypeMap
    return GenotypePhenotypeMap(gpm.wildtype,
                                d.genotypes,
                                phenotypes,
                                mutations=gpm.mutations,
                                stdeviations=stdeviations,
                                n_replicates=n_replicates)


def subset(gpm, **rows_to_select):
    """Create a subset GenotypePhenotypeMap from a given GenotypePhenotypeMap
    using an attribute of the Map object and a list of values.

    For example, we can subset a map using the genotypes attribute and give
    a list of genotypes to include in the subset

    .. code-block:: python

        gpm2 = subset(gpm, genotypes=['AA', 'AV'])


    Alternatively, we can subset the map using the binary attribute the same
    way.

    .. code-block:: python

        gpm2 = subset(gpm, binary=['00', '01'])
    """
    d = gpm.data
    items = list(rows_to_select.items())

    # Check kwargs is proper
    if len(items) > 1:
        raise Exception('This function can only call one column.')
    elif items[0][0] not in ['genotypes', 'binary', 'index']:
        raise Exception('Attribute given not in genotype-phenotype map.')
    elif items[0][0] == 'index':
        column = items[0][1]
        d_ = d.loc[items[0][1]]
    else:
        label = items[0][0]
        column = items[0][1]
        d_ = d.loc[d[label].isin(column)]

    # Not all the
    if len(d_) != len(column):
        raise Exception("Not all rows were found.")

    # Return GenotypePhenotypeMap
    return GenotypePhenotypeMap(gpm.wildtype,
                                d_.genotypes,
                                d_.phenotypes,
                                mutations=gpm.mutations,
                                stdeviations=d_.stdeviations,
                                n_replicates=d_.n_replicates)



# Mapping object for holding upper and lower error bars
#
# Author: Zach Sailer
#
# ----------------------------------------------------------
# Local imports
# ----------------------------------------------------------

import numpy as np


def upper_transform(mean, bound, logbase):
    """ Log transformation scaling.

    Examples
    --------
    Untransformed data looks as so:

        Yupper = Ymean + bound
        Ylower = Ymean - bound

    We want log(bounds)
        ie.
            log(Yupper) - log(Ymean)
            log(Ylower) + log(Ymean)

    so log(bound) = log(1 + bound/Ymean)
       log(bound) = log(1 - bound/Ymean)
    """
    return abs(logbase((mean + bound) / mean))


def lower_transform(mean, bound, logbase):
    """ Log transformation scaling.


    Examples
    --------
    Untransformed data looks as so:

        Yupper = Ymean + bound
        Ylower = Ymean - bound

    We want log(bounds)
        ie.
            log(Yupper) - log(Ymean)
            log(Ylower) + log(Ymean)

    so log(bound) = log(1 + bound/Ymean)
       log(bound) = log(1 - bound/Ymean)
    """
    return abs(logbase(mean / (mean - bound)))


class BaseErrorMap(object):
    """ Object to attach to seqspace objects for managing errors, standard
    deviations, and their log transforms.

    If a lower bound is given, use it instead of -variances.
    """

    def __init__(self, Map):
        self._Map = Map

    def wrapper(self, bound, **kwargs):
        """Wrapper function that changes variances to whatever bound desired.
        """
        raise Exception(""" Must be implemented in a subclass """)

    @property
    def upper(self):
        """Get upper error bound"""
        if self._Map.stdeviations is None:
            return None
        else:
            return self.wrapper(self._Map.stdeviations)

    @property
    def lower(self):
        """Get lower error bound."""
        if self._Map.stdeviations is None:
            return None
        else:
            return self.wrapper(self._Map.stdeviations)


class StandardDeviationMap(BaseErrorMap):

    def wrapper(self, bounds, **kwargs):
        """Wrapper function to convert Variances if necessary"""
        return bounds


class StandardErrorMap(BaseErrorMap):

    def wrapper(self, bounds):
        """Wrapper function to convert Variances if necessary"""
        return bounds / np.sqrt(self._Map.n_replicates)



import numpy as np
import random
from gpmap.gpm import GenotypePhenotypeMap
from gpmap import utils
from .base import random_mutation_set, BaseSimulation



class MultiPeakMountFujiSimulation(BaseSimulation):
    """Constructs a genotype-phenotype map from a Mount Fuji model. [1]_

    A Mount Fuji sets a "global" fitness peak (max) on a single genotype in
    the space. The fitness goes down as a function of hamming distance away
    from this genotype, called a "fitness field". The strength (or scale) of
    this field is linear and depends on the parameters `field_strength`.
    Roughness can be added to the Mount Fuji model using a random
    `roughness` parameter. This adds a random value (positive or negative)
    to each of the fitnesses. The multi-peak Mount Fuji model defines
    multiple fitness peaks across the map. Each of the peaks defines a
    fitness for every node as described above, but in the end, all but the
    highest fitness value will be discarded, i.e. the fitness of each node
    is defined by the nearest peak.


    .. math::

        f(g) = \\nu (g) + c \cdot d(g_0, g)

    where $\\nu$ is the roughness parameter, $c$ is the field strength,
    and $d$ is the hamming distance between genotype $g$ and the
    reference genotype.

    Parameters
    ----------
    wildtype : str
        reference genotype to put the

    mutations : dict
        mutations alphabet for each site

    field_strength : float
        field strength

    roughness_width : float
        Width of roughness distribution

    roughness_dist : str, 'normal'
        Distribution used to create noise around phenotypes.

    peak_n : int.
        The desired number of peaks in map. (Is only guaranteed if roughness_width is 0, else ruggedness might lead to
        additional local fitness peaks.)

    min_dist : int, None (default = 2).
        Minimal distance between any two peaks.

    max_dist : int, None (default = None).
        Maximal distance between any two peaks. If None, maximal distance is going to be the largest possible distance,
        i.e. 5 in a 5-site map.

    a_state : int, None (default = None).
        State that is guaranteed to become a peak. If None, a_state is going to be the wildtype genotype.
        This is done to have to well separated peaks that span the whole map. If unwanted, set any node as a_state
        and/or b_state.

    b_state : int, None (default = None).
        State that is guaranteed to become a peak. If None, b_state is going to be the farthest genotype from wildtype.
        This is done to have to well separated peaks that span the whole map. If unwanted, set any node as a_state
        and/or b_state.

    peaks : list, None (default = None).
        A list of nodes that will be defined as peaks. If None, peaks will be defined randomly using the other arguments
        as criteria.

    References
    ----------

    _ [1] Szendro, Ivan G., et al. "Quantitative analyses of empirical fitness
        landscapes." Journal of Statistical Mechanics: Theory and Experiment
        2013.01 (2013): P01005.
    """

    def __init__(
            self,
            wildtype,
            mutations,
            field_strength=1,
            roughness_width=None,
            roughness_dist='normal',
            peak_n=2,
            min_dist=1,
            max_dist=None,
            a_state=None,
            b_state=None,
            peaks=None,
            *args,
            **kwargs):
        # Call parent class.
        super(MultiPeakMountFujiSimulation, self).__init__(wildtype, mutations,
                                                  *args, **kwargs)
        # Set the field strength and roughness
        self._field_strength = field_strength
        self._roughness_width = roughness_width
        self._roughness_dist = roughness_dist
        self._roughness = None
        self._peak_n = peak_n
        self._peaks = peaks
        self._min_dist = min_dist
        self._max_dist = max_dist
        self._a_state = a_state
        self._b_state = b_state
        self.build()

    @classmethod
    def from_length(
        cls,
        length,
        field_strength=1,
        roughness_width=None,
        roughness_dist='normal',
        *args,
        **kwargs):
        """Constructs a genotype-phenotype map from a Mount Fuji model. [1]_

        A Mount Fuji sets a "global" fitness peak (max) on a single genotype in
        the space. The fitness goes down as a function of hamming distance away
        from this genotype, called a "fitness field". The strength (or scale) of
        this field is linear and depends on the parameters `field_strength`.
        Roughness can be added to the Mount Fuji model using a random
        `roughness` parameter. This assigns a random

        .. math::

            f(g) = \\nu (g) + c \cdot d(g_0, g)

        where $\\nu$ is the roughness parameter, $c$ is the field strength,
        and $d$ is the hamming distance between genotype $g$ and the
        reference genotype.

        Parameters
        ----------
        length : int
            length of the genotypes.

        field_strength : float
            field strength

        roughness_width : float
            Width of roughness distribution

        roughness_dist : str, 'normal'
            Distribution used to create noise around phenotypes.
        """
        cls = super(MultiPeakMountFujiSimulation, cls).from_length(
            length,
            field_strength=field_strength,
            roughness_width=roughness_width,
            roughness_dist=roughness_dist,
            *args,
            **kwargs
        )
        return cls

    @property
    def a_state(self):
        """Wild type state"""
        if self._a_state is not None:
            return self._a_state
        elif self._a_state is None:
            self._a_state = self.wildtype
            return self._a_state

    @property
    def b_state(self):
        """Derived state"""
        if self._b_state is not None:
            return self._b_state
        elif self._b_state is None:
            self._b_state = utils.farthest_genotype(self.a_state, self.genotypes)
            return self._b_state

    @property
    def peaks(self):
        if self._peaks:
            return self._peaks
        else:
            """Find n peaks that meet the max_dist/min_dist requirement"""
            self._peaks = [self.b_state, self.a_state]
            while len(self._peaks) < self.peak_n:
                proposed = random.choice(self.genotypes)  # Propose a new peak.
                add = False
                for peak in self._peaks:
                    dist = utils.hamming_distance(peak, proposed)
                    if dist >= self.min_dist and dist <= self.max_dist:  # Check dist. requirements
                        add = True
                    else:
                        add = False
                        break
                if add:
                    self._peaks.append(proposed)
            return self._peaks

    @property
    def hamming(self):
        """Hamming distances from each peak"""
        try:
            return self._hamming
        # calculate the hamming distance if not done already
        except AttributeError:
            hd = np.empty([len(self.peaks), len(self.genotypes)], dtype=int)
            for i, peak in enumerate(self.peaks):
                for j, g in enumerate(self.genotypes):
                    hd[i][j] = utils.hamming_distance(peak, g)
            self._hamming = hd
            return self._hamming

    @property
    def peak_n(self):
        """Number of peaks"""
        return self._peak_n

    @property
    def max_dist(self):
        """Maximum hamming distance between two peaks"""
        if self._max_dist is not None:
            return self._max_dist
        # If maximum distance between peaks is not given, set to the maximum hamming distance.
        elif self._max_dist is None:
            self._max_dist = len(list(self.genotypes[0]))
            return self._max_dist

    @property
    def min_dist(self):
        """Minimum hamming distance between two peaks"""
        return self._min_dist

    @property
    def roughness(self):
        """Array of roughness values for all genotypes"""
        if self._roughness is not None:
            return self._roughness

        elif self.roughness_width is None:
            return np.zeros(self.n)

        elif self.roughness_dist == 'normal':
            # Set roughness.
            self._roughness = np.random.normal(
                scale=self.roughness_width,
                size=self.n)

            return self._roughness

        elif self.roughness_dist == 'uniform':
            # Set roughness.
            self._roughness = np.random.uniform(
                high=self.roughness_width,
                low=-self.roughness_width,
                size=self.n)

            return self._roughness

        else:
            raise Exception("Roughness isn't set.")

    @property
    def roughness_dist(self):
        """Roughness distribution."""
        return self._roughness_dist

    @roughness_dist.setter
    def roughess_dist(self, roughness_dist):
        """Set the roughness distribution. Also sets the roughness array.s
        """
        # Make sure roughness dist is the right type.
        if not isinstance(roughness_dist, str):
            raise TypeError('roughness_dist must be a string.')

        # Get roughness distribution
        if roughness_dist not in ['normal', 'uniform']:
            raise AttributeError('roughness_dist must be '
                                 'either normal or uniform')

        # Set roughness distribution and reset map
        self._roughness_dist = roughness_dist
        self._roughness = None
        self.build()

    @property
    def roughness_width(self):
        return self._roughness_width

    @roughness_width.setter
    def roughness_width(self, roughness_width):
        # Set roughness distribution and reset map
        self._roughness_width = roughness_width
        self._roughness = None
        self.build()

    @property
    def field_strength(self):
        return self._field_strength

    @field_strength.setter
    def field_strength(self, c):
        self._field_strength = c
        self.build()

    @property
    def scale(self):
        """Multipeak Mt. Fuji phenotypes without noise."""
        hd = np.empty([len(self.peaks), len(self.genotypes)])
        for i, peak in enumerate(self.peaks):
            newrow = self.hamming[i] * self.field_strength
            hd[i] = newrow

        min_hd = hd.min(0)  # Column-wise minimum value of array.
        max_min = np.amax(min_hd)  # Get the maximum value of the array for normalization.
        self._scale = 1 - (min_hd / max_min)  # Subtract from one -> Larger hamming dist. from peak = lower phenotype.
        return self._scale

    def build(self):
        """Construct phenotypes using a rough Mount Fuji model."""
        self.data.phenotypes = self.roughness + self.scale



# Import utils model.
from .. import utils

WILDTYPE = "AAA"

GENOTYPES = [
    "AAA",
    "AAB",
    "ABA",
    "BAA",
    "ABB",
    "BAB",
    "BBA",
    "BBB"
]

BINARY = [
    '000',
    '001',
    '010',
    '100',
    '011',
    '101',
    '110',
    '111'
]

MUTATIONS = {
    0: ["A", "B"],
    1: ["A", "B"],
    2: ["A", "B"],
}

def lists_are_same(list1, list2):
    """Return true if lists contain same values
    (order doesn't matter).
    """
    set1, set2 = set(list1), set(list2)

    # Find differences
    if set1.issubset(set2) and set1.issuperset(set2):
        return True

    else:
        return False


def dicts_are_same(dict1, dict2):
    """Return True if dictionaries are the same.
    """
    # Check that keys are the same.
    keys1, keys2 = list(dict1.keys()), list(dict2.keys())
    if not lists_are_same(keys1, keys2):
        return False

    # Check that values are the same.
    for key, val in dict1.items():
        if dict2[key] != val:
            return False

    return True

def test_hamming_distance():
    """Test hamming distance function."""
    s1 = "THIS IS A TEST"
    s2 = "HWIS IT A TEBT"

    assert utils.hamming_distance(s1, s2) == 4


def test_find_differences():
    """Test find differences function."""
    s1 = "THIS IS A TEST"
    s2 = "HWIS IT A TEBT"

    assert utils.find_differences(s1, s2) == [0,1,6,12]


def test_farthest_genotype():
    """Test farthest genotypes function."""
    assert utils.farthest_genotype("AAA", GENOTYPES) == "BBB"


def test_list_binary():
    """test list binary function."""
    bin = utils.list_binary(3)

    assert lists_are_same(bin, BINARY)


def test_mutations_to_encoding():
    """Test mutations to encoding."""
    encoding = utils.mutations_to_encoding(WILDTYPE, MUTATIONS)

    assert encoding[0]['A'] == '0'
    assert encoding[1]['B'] == '1'
    assert encoding[2]['A'] == '0'


def test_mutations_to_genotypes():
    """Test mutations to genotypes function."""
    genotypes = utils.mutations_to_genotypes(MUTATIONS, wildtype=WILDTYPE)

    assert lists_are_same(genotypes, GENOTYPES)


def test_genotypes_to_mutations():
    """Test genotypes to mutations function."""
    mutations = utils.genotypes_to_mutations(GENOTYPES)

    assert dicts_are_same(mutations, MUTATIONS)


def test_genotypes_to_binary():
    """Test get_missing_genotypes function."""
    encoding_table = utils.get_encoding_table(WILDTYPE, MUTATIONS)
    binary = utils.genotypes_to_binary(GENOTYPES, encoding_table)

    assert lists_are_same(binary, BINARY)


def test_get_missing_genotypes():
    """Test get_missing_genotypes function."""
    known_, missing_ = GENOTYPES[0:4], GENOTYPES[4:]

    missing = utils.get_missing_genotypes(known_, MUTATIONS)

    assert lists_are_same(missing, missing_)



import random
import numpy as np
from gpmap import utils
from gpmap.gpm import GenotypePhenotypeMap


def random_mutation_set(length, alphabet_size=2, type='AA'):
    """Generate a random mutations dictionary for simulations.

    Parameters
    ----------
    length : length of genotypes

    alphabet_size : int or list
        alphabet size at each site. if list is given, will make site i have
        size alphab_size[i].
    type : 'AA' or "DNA'
        Use amino acid alphabet or DNA alphabet
    """

    if isinstance(alphabet_size, int):
        size = [alphabet_size for i in range(length)]
    else:
        size = alphabet_size
    # build mutations dictionary
    mutations = {}
    for i in range(length):
        if type == 'AA':
            A = utils.AMINO_ACIDS
        elif type == 'DNA':
            A = utils.DNA
        random.shuffle(A)
        alphabet = A[:size[i]]
        mutations[i] = alphabet
    return mutations


class BaseSimulation(GenotypePhenotypeMap):
    """ Build a simulated GenotypePhenotypeMap. Generates random phenotypes.
    """

    def __init__(self, wildtype, mutations, *args, **kwargs):
        # build genotypes
        genotypes = utils.mutations_to_genotypes(mutations, wildtype=wildtype)
        phenotypes = np.empty(len(genotypes), dtype=float)
        super(BaseSimulation, self).__init__(wildtype,
                                             genotypes,
                                             phenotypes,
                                             mutations=mutations,
                                             *args,
                                             **kwargs)

    @classmethod
    def from_length(cls, length, alphabet_size=2, *args, **kwargs):
        """ Create a simulate genotype-phenotype map from a given genotype length.

        Parameters
        ----------
        length : int
            length of genotypes
        alphabet_size : int (optional)
            alphabet size

        Returns
        -------
        self : GenotypePhenotypeSimulation
        """
        mutations = random_mutation_set(length, alphabet_size=alphabet_size)
        wildtype = "".join([m[0] for m in mutations.values()])
        self = cls(wildtype, mutations, *args, **kwargs)
        return self

    def set_stdeviations(self, sigma):
        """Add standard deviations to the simulated phenotypes, which can then be
        used for sampling error in the genotype-phenotype map.

        Parameters
        ----------
        sigma : float or array-like
            Adds standard deviations to the phenotypes. If float, all
            phenotypes are given the same stdeviations. Else, array must be
            same length as phenotypes and will be assigned to each phenotype.
        """
        stdeviations = np.ones(len(self.phenotypes)) * sigma
        self.data.stdeviations = stdeviations
        return self

    def build(self):
        raise Exception("must be implemented in subclass.")



__doc__ = """Utility functions for managing genotype-phenotype map data
and conversions.

Glossary:
--------
mutations : doct
    keys are site numbers in the genotypes. Values are alphabet of mutations at
    that sites

encoding : dict
    keys are site numbers in genotype. Values are dictionaries mapping each
    mutation to its binary representation.



"""

# -------------------------------------------------------
# Miscellaneous Python functions for random task
# -------------------------------------------------------

import itertools as it
import numpy as np
from scipy.misc import comb
from collections import OrderedDict
import warnings
import pandas as pd

# -------------------------------------------------------
# Mutation alphabets
# -------------------------------------------------------

DNA = ["A", "C", "G", "T"]

AMINO_ACIDS = ["D", "T", "S", "E", "P", "G", "A", "C", "V", "M", "I",
               "L", "Y", "F", "H", "K", "R", "W", "Q", "N"]

# -------------------------------------------------------
# Wrappers for methods that use optional imports
# -------------------------------------------------------


def ipywidgets_missing(function):
    """Wrapper checks that ipython widgets are install before trying to
    render them.
    """
    def wrapper(*args, **kwargs):
        try:

            import ipywidgets
            return function(*args, **kwargs)

        except ImportError:
            warnings.filterwarnings("once")
            warnings.warn(
                """Looks like `ipywidgets` is not installed, so widgets can't "
                "be constructed. Install before using this method.""",
                ImportWarning)

    return wrapper

# -------------------------------------------------------
# Useful methods for genotype-phenotype spaces
# -------------------------------------------------------


def get_base(logbase):
    """Get base from logbase
    Parameters
    ----------
    logbase : callable
        logarithm function
    Returns
    -------
    base : float
        returns base of logarithm.
    """
    testval = 10
    return np.exp(np.log(testval) / logbase(testval))


def hamming_distance(s1, s2):
    """Return the Hamming distance between equal-length sequences """
    return sum(ch1 != ch2 for ch1, ch2 in zip(s1, s2))


def sample_phenotypes(phenotypes, errors, n=1):
    """Generate `n` phenotypes from from normal distributions. """
    samples = np.random.randn(len(phenotypes), n)
    # Apply phenotype scale and variance
    for i in range(n):
        samples[:, i] = np.multiply(samples[:, i], errors) + phenotypes
    return samples

# -------------------------------------------------------
# Utilities for searching sequence space
# -------------------------------------------------------


def find_differences(s1, s2):
    """Return the index of differences between two sequences."""
    indices = list()
    for i in range(len(s1)):
        if s1[i] != s2[i]:
            indices.append(i)
    return indices


def farthest_genotype(reference, genotypes):
    """Find the genotype in the system that differs at the most sites. """
    mutations = 0
    for genotype in genotypes:
        differs = hamming_distance(genotype, reference)
        if differs > mutations:
            mutations = int(differs)
            mutant = str(genotype)
    return mutant

# -------------------------------------------------------
# Space enumerations
# -------------------------------------------------------


def list_binary(length):
    """List all binary strings with given length.
    """
    return np.array(["".join(seq) for seq in it.product("01", repeat=length)])


def get_encoding_table(wildtype, mutations):
    """This function constructs a lookup table (pandas.DataFrame) for mutations
    in a given mutations dictionary. This table encodes mutations with a binary representation.
    """
    # Initialize table
    table = []
    mutation_index_counter = 0
    binary_index_counter = 0
    for genotype_index, alphabet in mutations.items():
        # Type check genotype_index
        genotype_index = int(genotype_index)

        # Handle sites that don't mutate.
        if alphabet is None:
            # Create a row for the encoding lookup table.
            table.append(dict(
                genotype_index=genotype_index,
                wildtype_letter=wildtype[genotype_index],
                mutation_letter=None,
                binary_repr="",
                binary_index_start=binary_index,
                binary_index_stop=binary_index,
                mutation_index=None
            ))

        # Determine mapping for all other sites.
        else:
            # copy alphabet to avoid removing items in main object.
            alphabet_cp = alphabet[:]
            n = len(alphabet_cp) - 1  # number of mutation neighbors
            binary_index = binary_index_counter

            # Set wildtype state at a given genotype_index.
            wt_site = wildtype[genotype_index]
            table.append(dict(
                genotype_index=genotype_index,
                wildtype_letter=wt_site,
                mutation_letter=wt_site,
                binary_repr="0" * n,
                binary_index_start=binary_index,
                binary_index_stop=binary_index+n,
                mutation_index=None
            ))

            # Copy alphabet again to prevent indexing error.
            alphabet_ = alphabet_cp[:]
            alphabet_.remove(wt_site)

            # Add all possible mutations at given site 
            for j in range(n):
                binary_repr = list("0" * n)
                binary_repr[j] = "1"
                binary_repr = "".join(binary_repr)
                table.append(dict(
                    genotype_index=genotype_index,
                    wildtype_letter=wt_site,
                    mutation_letter=alphabet_[j],
                    binary_repr=binary_repr,
                    binary_index_start=binary_index,
                    binary_index_stop=binary_index+n,
                    mutation_index=mutation_index_counter + 1
                ))
                mutation_index_counter += 1
            binary_index_counter += n

    # Turn table into DataFrame.    
    df = pd.DataFrame(table)
    df.genotype_index = df.genotype_index.astype('Int64')
    df.mutation_index = df.mutation_index.astype('Int64')
    df.binary_index_start = df.binary_index_start.astype('Int64')
    df.binary_index_stop = df.binary_index_stop.astype('Int64')
    return df


def genotypes_to_binary(genotypes, encoding_table):
    """Using an encoding table (see `get_encoding_table` 
    function), build a set of binary genotypes.

    Parameters
    ----------
    genotypes : 
        List of the genotypes to encode.
    encoding_table : 
        DataFrame that encodes the binary representation of
        each mutation in the list of genotypes. (See the 
        `get_encoding_table`).
    """
    # ---------- Sanity Checks ---------------
    # 1. Check genotypes are all same length
    length_of_genotypes = [len(g) for g in genotypes]
    length = length_of_genotypes[0]

    if len(set(length_of_genotypes)) > 1:
        raise Exception("Genotypes are not all the same length.")
        
    binary = []
    # Alias for encoding table
    t = encoding_table
    mapper = dict(zip(zip(t.genotype_index, t.mutation_letter), t.binary_repr))
    for g in genotypes:
        b = []
        for genotype_index, mutation_letter in enumerate(g):
            chunk = mapper[(genotype_index, mutation_letter)]
            b.append(chunk)
        binary.append("".join(b))
    return binary


def mutations_to_encoding(wildtype, mutations):
    """ Encoding map for genotype-to-binary

    Parameters
    ---------
    wildtype: str
        Wildtype sequence.
    mutations: dict
        Mapping of each site's mutation alphabet.
        {site-number: [alphabet]}

    Returns
    -------
    encode: OrderedDict of OrderDicts
        Encoding dictionary that maps site number to mutation-binary map


    Examples
    --------
    ``{ <site-number> : {<mutation>: <binary>} }``
    """
    encoding = OrderedDict()

    for site_number, alphabet in mutations.items():
        site_number = int(site_number)
        # Handle sites that don't mutate.
        if alphabet is None:
            encoding[site_number] = wildtype[site_number]

        # All sites that mutate, give a mapping dictionary.
        else:
            # copy alphabet to avoid removing items in main object.
            alphabet_cp = alphabet[:]
            n = len(alphabet_cp) - 1  # number of mutation neighbors
            wt_site = wildtype[site_number]  # wildtype letter

            # Build a binary representation of mutation alphabet
            indiv_encode = OrderedDict({wt_site: "0" * n})
            alphabet_ = alphabet_cp[:]
            alphabet_.remove(wt_site)

            for i in range(n):
                binary = list("0" * n)
                binary[i] = "1"
                indiv_encode[alphabet_[i]] = "".join(binary)
            encoding[site_number] = indiv_encode

    return encoding


def mutations_to_genotypes(mutations, wildtype=None):
    """Use a mutations dictionary to construct an array of genotypes composed
    of those mutations.

    Parameters
    ----------
    mutations : dict
        A mapping dict with site numbers as keys and lists of mutations as
        values.

    wildtype : str
        wildtype genotype (as string).

    Returns
    -------
    genotypes : list
        list of genotypes comprised of mutations in given dictionary.
    """
    # Convert mutations dict to list of lists
    mutations_ = []
    for i, val in enumerate(mutations.values()):
        if val is None:
            mutations_.append(wildtype[i])
        else:
            mutations_.append(val)
    sequences = it.product(*mutations_)
    genotypes = ["".join(s) for s in sequences]
    return genotypes


def genotypes_to_mutations(genotypes):
    """Create mutations dictionary from a list of mutations.
    """
    # Sequences to array
    arr = np.array([list(g) for g in genotypes])

    # Mutations dict
    mutations = {i: None for i in range(arr.shape[1])}

    # Find unique residues at all sites.
    for i, col in enumerate(arr.T):
        mutations[i] = list(np.unique(col))

    return mutations


def get_missing_genotypes(genotypes, mutations=None):
    """Get a list of genotypes not found in the given genotypes list.

    Parameters
    ----------
    genotypes : list
        List of genotypes.

    mutations : dict (optional)
        Mutation dictionary

    Return
    ------
    missing_genotypes : list
        List of genotypes not found in genotypes list.
    """
    if mutations is None:
        mutations = genotypes_to_mutations(genotypes)

    # Need a wildtype--doesn't matter what it is.
    wildtype = "".join([sites[0] for sites in mutations.values()])

    # Get all genotypes.
    all_genotypes = mutations_to_genotypes(mutations, wildtype)

    # Find genotypes not found in genotypes list.
    missing_genotypes = set(all_genotypes).difference(set(genotypes))
    return list(missing_genotypes)

def length_to_mutations(length, alphabet=["0", "1"]):
    """Build a mutations dictionary for a given alphabet
    
    Parameters
    ----------
    length : int
        length of the genotypes

    alphabet : list
        List of mutations at each site.  
    """
    return {i: alphabet for i in range(length)}


#
# Author: Zach Sailer
#
# ----------------------------------------------------------
# Outside imports
# ----------------------------------------------------------

import json
import pickle
import numpy as np
import pandas as pd

# ----------------------------------------------------------
# Local imports
# ----------------------------------------------------------

# import different maps into this module
import gpmap.utils as utils
import gpmap.errors as errors


class GenotypePhenotypeMap(object):
    """Object for containing genotype-phenotype map data.

    Parameters
    ----------
    wildtype : string
        wildtype sequence.

    genotypes : array-like
        list of all genotypes in system. Must be a complete system.

    phenotypes : array-like
        List of phenotypes in the same order as genotypes.

    mutations : dict
        Dictionary that maps each site indice to their possible substitution
        alphabet.

    n_replicates : int
        number of replicate measurements comprising the mean phenotypes

    include_binary : bool (default=True)
        Construct a binary representation of the space.

    Attributes
    ----------
    data : pandas.DataFrame
        The core data object. Columns are 'genotypes', 'phenotypes',
        'n_replicates', 'stdeviations', and (option) 'binary'.

    complete_data : pandas.DataFrame (optional, created by BinaryMap)
        A dataframe mapping the complete set of genotypes possible, given
        the mutations dictionary. Contains all columns in `data`. Any missing
        data is reported as NaN.

    missing_data : pandas.DataFrame (optional, created by BinaryMap)
        A dataframe containing the set of missing genotypes; complte_data -
        data. Two columns: 'genotypes' and 'binary'.

    binary : BinaryMap
        object that gives you (the user) access to the binary representation
        of the map.

    encoding_table:
        Pandas DataFrame showing how mutations map to binary representation.
    """
    def __init__(self, wildtype, genotypes, phenotypes,
                 stdeviations=None,
                 mutations=None,
                 n_replicates=1,
                 **kwargs):

        # Set mutations; if not given, assume binary space.
        if mutations is not None:
            # Make sure the keys in the mutations dict are integers, not
            # strings.
            self._mutations = dict([(int(key), val)
                                   for key, val in mutations.items()])
        else:
            # Get mutations dict from genotypes.
            mutations = utils.genotypes_to_mutations(genotypes)
            self._mutations = mutations

        # Leftover kwargs become metadata that is ignored.
        self.metadata = kwargs

        # Set wildtype.
        self._wildtype = wildtype

        # Store data in DataFrame
        data = dict(
            genotypes=genotypes,
            phenotypes=phenotypes,
            n_replicates=n_replicates,
            stdeviations=stdeviations
        )
        self.data = pd.DataFrame(data)
        
        # Construct a lookup table for all mutations.
        self.encoding_table = utils.get_encoding_table(
            self.wildtype,
            self.mutations
        )

        # Add binary representation
        self.add_binary()

        # Construct the error maps
        self._add_error()

    def _repr_html_(self):
        """Represent the GenotypePhenotypeMap as an html table."""
        return self.data.to_html()

    def map(self, attr1, attr2):
        """Dictionary that maps attr1 to attr2."""
        return dict(zip(getattr(self, attr1), getattr(self, attr2)))

    @classmethod
    def read_dataframe(cls, dataframe, wildtype, **kwargs):
        """Construct a GenotypePhenotypeMap from a dataframe."""
        # Required arguments
        df = dataframe
        self = cls(wildtype,
                   df.genotypes,
                   df.phenotypes,
                   stdeviations=df.stdeviations,
                   n_replicates=df.n_replicates,
                   **kwargs)
        return self

    @classmethod
    def read_pickle(cls, filename, **kwargs):
        """Read GenotypePhenotypeMap from pickle"""
        with open(filename, 'rb') as f:
            self = pickle.load(f)

        # Check input
        if not isinstance(self, GenotypePhenotypeMap):
            raise Exception("Pickle file does not contain a GenotypePhenotypeMap.")

        return self

    @classmethod
    def read_csv(cls, fname, wildtype, **kwargs):
        """"""
        dtypes = dict(
            genotypes=str,
            phenotypes=float,
            stdeviations=float,
            n_replicates=int
        )
        df = pd.read_csv(fname, dtype=dtypes)
        self = cls.read_dataframe(df, wildtype, **kwargs)
        return self

    @classmethod
    def read_excel(cls, fname, wildtype, **kwargs):
        """"""
        dtypes = dict(
            genotypes=str,
            phenotypes=float,
            stdeviations=float,
            n_replicates=int
        )
        df = pd.read_excel(fname, dtype=dtypes)
        self = cls.read_dataframe(df, wildtype, **kwargs)
        return self

    @classmethod
    def read_json(cls, filename, **kwargs):
        """Load a genotype-phenotype map directly from a json file.
        The JSON metadata must include the following attributes

        Note
        ----
        Keyword arguments override input that is loaded from the JSON file.
        """
        # Open, json load, and close a json file
        with open(filename, "r") as f:        
            metadata = json.load(f)

        return cls.from_dict(metadata)


    @classmethod
    def from_dict(cls, metadata):
        """"""
        data = metadata["data"]

        if "wildtype" in metadata:
            wildtype = metadata["wildtype"]
            metadata.pop("wildtype")

        # Check keys in dictionary.
        if not all(key in data for key in ["genotypes", "phenotypes", "stdeviations", "n_replicates"]):
            raise Exception('The "data" field must have the following keys: '
                            'genotypes", "phenotypes", "stdeviations", "n_replicates"')

        # Create an instance
        gpm = cls(
            wildtype,
            data["genotypes"],
            data["phenotypes"],
            stdeviations=data["stdeviations"],
            n_replicates=data["n_replicates"],
            **metadata
        )
        return gpm

    @classmethod
    def from_json(cls, json_str):
        """Load a genotype-phenotype map directly from a json.
        The JSON metadata must include the following attributes

        Note
        ----
        Keyword arguments override input that is loaded from the JSON file.
        """
        metadata = json.loads(json_str)
        return cls.from_dict(metadata)

    # ----------------------------------------------------------
    # Writing methods
    # ----------------------------------------------------------

    def to_pickle(self, filename, **kwargs):
        """Write GenotypePhenotypeMap object to a pickle file.
        """
        with open(filename, 'wb') as f:
            pickle.dump(self, f)

    def to_excel(self, filename=None, **kwargs):
        """Write genotype-phenotype map to excel spreadsheet.

        Keyword arguments are passed directly to Pandas dataframe to_excel
        method.

        Parameters
        ----------
        filename : str
            Name of file to write out.
        """
        self.data.to_excel(filename, **kwargs)

    def to_csv(self, filename=None, **kwargs):
        """Write genotype-phenotype map to csv spreadsheet.

        Keyword arguments are passed directly to Pandas dataframe to_csv
        method.

        Parameters
        ----------
        filename : str
            Name of file to write out.
        """
        self.data.to_csv(filename, **kwargs)

    def to_dict(self, complete=False):
        """Write genotype-phenotype map to dict."""
        if complete:
            data = self.complete_data.to_dict('list')
        else:
            data = self.data.to_dict('list')

        metadata = {
            "wildtype": self.wildtype,
            "mutations": self.mutations,
            "data": data
        }
        metadata.update(**self.metadata)
        return metadata

    def to_json(self, filename=None, complete=False):
        """Write genotype-phenotype map to json file. If no filename is given
        returns
        """
        # Get metadata.
        data = self.to_dict(complete=complete)

        # Write to json file.
        if filename is None:
            return json.dumps(data)
        else:
            with open(filename, "w") as f:
                json.dump(data, f)

    @property
    def length(self):
        """Get length of the genotypes. """
        return len(self.wildtype)

    @property
    def n(self):
        """Get number of genotypes, i.e. size of the genotype-phenotype map."""
        return len(self.genotypes)

    @property
    def wildtype(self):
        """Get reference genotypes for interactions. """
        return self._wildtype

    @wildtype.setter
    def wildtype(self, wildtype):
        """If a wildtype is given after init, rebuild binary genotypes."""
        self._wildtype = wildtype
        self.add_binary()

    @property
    def mutant(self):
        """Get the farthest mutant in genotype-phenotype map."""
        _mutant = []
        _wt = self.wildtype
        for i in range(0, len(self.mutations)):
            site = _wt[i]
            options = self.mutations[i]
            if options is None:
                _mutant.append(_wt[i])
            else:
                for o in options:
                    if o != site:
                        _mutant.append(o)
        return "".join(_mutant)

    @property
    def mutations(self):
        """Get the furthest genotype from the wildtype genotype."""
        return self._mutations

    @property
    def genotypes(self):
        """Get the genotypes of the system."""
        return self.data.genotypes.values

    @property
    def binary(self):
        """Binary representation of genotypes."""
        return self.data.binary.values

    @property
    def phenotypes(self):
        """Get the phenotypes of the system. """
        return self.data.phenotypes.values

    @property
    def stdeviations(self):
        """Get stdeviations"""
        return self.data.stdeviations.values

    @property
    def n_replicates(self):
        """Return the number of replicate measurements made of the phenotype"""
        return self.data.n_replicates.values

    @property
    def index(self):
        """Return numpy array of genotypes position. """
        return self.data.index.values

    def _add_error(self):
        """Store error maps"""
        self.std = errors.StandardDeviationMap(self)
        self.err = errors.StandardErrorMap(self)

    def add_binary(self):
        """Build a binary representation of set of genotypes.
        """
        binary = utils.genotypes_to_binary(self.genotypes, self.encoding_table)

        # Add this as a column to the map.
        self.data['binary'] = binary

    def get_missing_genotypes(self):
        """Get all genotypes missing from the complete genotype-phenotype map."""
        return utils.get_missing_genotypes(
            self.genotypes,
            mutations=self.mutations
        )

    def get_all_possible_genotypes(self):
        """Get the complete set of genotypes possible. There is no particular order
        to the genotypes. Consider sorting.
        """
        # Get all genotypes.
        return utils.mutations_to_genotypes(self.mutations, wildtype=self.wildtype)


# import numpy as np
# import pytest
#
# from ..gpm import GenotypePhenotypeMap
#
# @pytest.fixture()
# def tmp_gpm_file():
#     pass
#
#
# class testGenotypePhenotypeMap:
#
#     def _test_attributes(self, gpm):
#         # Test instance was created
#         assert isinstance(gpm, GenotypePhenotypeMap)
#         # Test elements align
#         np.testing.assert_array_equal(gpm.genotypes, self.genotypes)
#
#     def test_init(self):
#         """Test initialization of class"""
#         gpm = GenotypePhenotypeMap(
#             self.wildtype,
#             self.genotypes,
#             self.phenotypes,
#             n_replicates=self.n_replicates,
#             mutations=self.mutations
#         )
#         # Test instance was created
#         tools.assert_is_instance(gpm, GenotypePhenotypeMap)
#         # Test elements align
#         np.testing.assert_array_equal(gpm.genotypes, self.genotypes)
#         np.testing.assert_array_equal(gpm.phenotypes, self.phenotypes)
#
#     def test_read_json(self):
#         """Test reading from json"""
#         # gpm = GenotypePhenotypeMap.read_json("data.json")
#         # Test instance was created
#         tools.assert_is_instance(gpm, GenotypePhenotypeMap)
#         # Test elements align
#         np.testing.assert_array_equal(gpm.genotypes, self.genotypes)
#
#     def test_length(self):
#         """Test genotype length."""
#         gpm = GenotypePhenotypeMap.read_json("data.json")
#         tools.assert_equal(self.GPM.length, 4)
#
#     def test_n(self):
#         """Test size"""
#         tools.assert_equal(self.GPM.n, 2**4)
#
#     def test_std(self):
#         """Test raw errors"""
#         np.testing.assert_array_equal(self.GPM.std.upper, self.stdeviations)
#
#     def test_mutant(self):
#         """Test mutant"""
#         tools.assert_equal(self.GPM.mutant, "1111")
#
#     def test_missing_genotypes(self):
#         """Test that missing genotypes are identified."""
#         # Choose a subset indices for testing missing genotypes.
#         index = np.arange(0,16)
#         np.random.shuffle(index)
#         # Get subset genotypes
#         chosen_g = np.array(self.genotypes)[index[:8]]
#         missing_g = np.array(self.genotypes)[index[8:]]
#
#         # Get subset phenotypes
#         chosen_p = np.array(self.phenotypes)[index[:8]]
#         # Init GPM
#         gpm = GenotypePhenotypeMap(
#             self.wildtype,
#             chosen_g,
#             chosen_p,
#             log_transform=self.log_transform,
#             n_replicates=self.n_replicates,
#             mutations=self.mutations,
#         )
#         # Test missing genotypes are returned
#         np.testing.assert_array_equal(gpm.genotypes, chosen_g)
#         np.testing.assert_array_equal(np.sort(gpm.missing_genotypes), np.sort(missing_g))



import numpy as np
from gpmap.gpm import GenotypePhenotypeMap
from gpmap import utils
from .base import random_mutation_set, BaseSimulation


class MountFujiSimulation(BaseSimulation):
    """Constructs a genotype-phenotype map from a Mount Fuji model. [1]_

    A Mount Fuji sets a "global" fitness peak (max) on a single genotype in
    the space. The fitness goes down as a function of hamming distance away
    from this genotype, called a "fitness field". The strength (or scale) of
    this field is linear and depends on the parameters `field_strength`.
    Roughness can be added to the Mount Fuji model using a random
    `roughness` parameter. This assigns a random

    .. math::

        f(g) = \\nu (g) + c \cdot d(g_0, g)

    where $\\nu$ is the roughness parameter, $c$ is the field strength,
    and $d$ is the hamming distance between genotype $g$ and the
    reference genotype.

    Parameters
    ----------
    wildtype : str
        reference genotype to put the

    mutations : dict
        mutations alphabet for each site

    field_strength : float
        field strength

    roughness_width : float
        Width of roughness distribution

    roughness_dist : str, 'normal'
        Distribution used to create noise around phenotypes.


    References
    ----------

    _ [1] Szendro, Ivan G., et al. "Quantitative analyses of empirical fitness
        landscapes." Journal of Statistical Mechanics: Theory and Experiment
        2013.01 (2013): P01005.
    """

    def __init__(
            self,
            wildtype,
            mutations,
            field_strength=1,
            roughness_width=None,
            roughness_dist='normal',
            *args,
            **kwargs):
        # Call parent class.
        super(MountFujiSimulation, self).__init__(wildtype, mutations,
                                                  *args, **kwargs)
        # Set the field strength and roughness
        self._field_strength = field_strength
        self._roughness_width = roughness_width
        self._roughness_dist = roughness_dist
        self._roughness = None
        self.build()

    @classmethod
    def from_length(
        cls,
        length,
        field_strength=1,
        roughness_width=None,
        roughness_dist='normal',
        *args,
        **kwargs):
        """Constructs a genotype-phenotype map from a Mount Fuji model. [1]_

        A Mount Fuji sets a "global" fitness peak (max) on a single genotype in
        the space. The fitness goes down as a function of hamming distance away
        from this genotype, called a "fitness field". The strength (or scale) of
        this field is linear and depends on the parameters `field_strength`.
        Roughness can be added to the Mount Fuji model using a random
        `roughness` parameter. This assigns a random

        .. math::

            f(g) = \\nu (g) + c \cdot d(g_0, g)

        where $\\nu$ is the roughness parameter, $c$ is the field strength,
        and $d$ is the hamming distance between genotype $g$ and the
        reference genotype.

        Parameters
        ----------
        length : int
            length of the genotypes.

        field_strength : float
            field strength

        roughness_width : float
            Width of roughness distribution

        roughness_dist : str, 'normal'
            Distribution used to create noise around phenotypes.
        """
        cls = super(MountFujiSimulation, cls).from_length(
            length,
            field_strength=field_strength,
            roughness_width=roughness_width,
            roughness_dist=roughness_dist,
            *args,
            **kwargs
        )
        return cls

    @property
    def hamming(self):
        """Hamming distance from reference"""
        try:
            return self._hamming
        # calculate the hamming distance if not done already
        except AttributeError:
            hd = np.empty(self.n, dtype=int)
            for i, g in enumerate(self.genotypes):
                hd[i] = utils.hamming_distance(self.wildtype, g)
            self._hamming = hd
            return self._hamming

    @property
    def roughness(self):
        """Array of roughness values for all genotypes"""
        if self._roughness is not None:
            return self._roughness

        elif self.roughness_width is None:
            return np.zeros(self.n)

        elif self.roughness_dist == 'normal':
            # Set roughness.
            self._roughness = np.random.normal(
                scale=self.roughness_width,
                size=self.n)

            return self._roughness

        elif self.roughness_dist == 'uniform':
            # Set roughness.
            self._roughness = np.random.uniform(
                high=self.roughness_width,
                low=-self.roughness_width,
                size=self.n)

            return self._roughness

        else:
            raise Exception("Roughness isn't set.")

    @property
    def roughness_dist(self):
        """Roughness distribution."""
        return self._roughness_dist

    @roughness_dist.setter
    def roughess_dist(self, roughness_dist):
        """Set the roughness distribution. Also sets the roughness array.s
        """
        # Make sure roughness dist is the right type.
        if not isinstance(roughness_dist, str):
            raise TypeError('roughness_dist must be a string.')

        # Get roughness distribution
        if roughness_dist not in ['normal', 'uniform']:
            raise AttributeError('roughness_dist must be '
                                 'either normal or uniform')

        # Set roughness distribution and reset map
        self._roughness_dist = roughness_dist
        self._roughness = None
        self.build()

    @property
    def roughness_width(self):
        return self._roughness_width

    @roughness_width.setter
    def roughness_width(self, roughness_width):
        # Set roughness distribution and reset map
        self._roughness_width = roughness_width
        self._roughness = None
        self.build()

    @property
    def field_strength(self):
        return self._field_strength

    @field_strength.setter
    def field_strength(self, c):
        self._field_strength = c
        self.build()

    @property
    def scale(self):
        """Mt. Fuji phenotypes without noise."""
        return self.field_strength * self.hamming

    def build(self):
        """Construct phenotypes using a rough Mount Fuji model."""
        self.data.phenotypes = self.roughness + self.scale



import numpy as np
from .base import BaseSimulation


class RandomPhenotypesSimulation(BaseSimulation):
    """ Build a simulated GenotypePhenotypeMap. Generates random phenotypes.
    """

    def __init__(self, wildtype, mutations, phenotype_range=(0, 1),
                 *args, **kwargs):
        # build genotypes
        super(RandomPhenotypesSimulation, self).__init__(wildtype,
                                                         mutations,
                                                         *args, **kwargs)
        self.phenotype_range = phenotype_range
        self.build()

    def build(self):
        """Build phenotypes"""
        low, high = self.phenotype_range[0], self.phenotype_range[1]
        self.data['phenotypes'] = np.random.uniform(low, high,
                                                    size=len(self.genotypes))



import numpy as np
import math

# -----------------------------------------------------------------------
# Unbiased calculations of sample statistics to error statistics
# -----------------------------------------------------------------------

def coverage(gpm):
    # Get binary representation
    binary = gpm.binary

    # Determine shape of obs matrix
    shape = (len(binary), len(binary[0]))

    # Initialize obs matrix
    obs_matrix = np.array([np.array(list(b)) for b in binary], dtype=int)

    # Compute the average times each mutation is observed
    observations = obs_matrix.sum(axis=0)



def c4_correction(n_samples):
    """Return the correction scalar for calculating standard deviation from a normal distribution. """
    k1 = round(n_samples / 2.0, 4)
    k2 = round((n_samples - 1) / 2.0, 4)

    # If the number of samples is < 100, calculate a correction scalar.
    if n_samples < 100:

        if k1.is_integer():
            c4 = np.sqrt(2.0 / (math.pi * (2 * k1 - 1))) * ((2**(2 * k1 - 2) *
                                                             math.factorial(k1 - 1)**2) / math.factorial(2 * k1 - 2))
        elif k2.is_integer():
            c4 = np.sqrt(math.pi / k2) * (math.factorial(2 * k2 - 1) /
                                          (2**(2 * k2 - 1) * math.factorial(k2 - 1)**2))
        else:
            raise Exception("""Non-integer value for correction term, c4.""")

    # Else this scalar is 1
    else:
        c4 = 1.0

    return c4


def unbiased_var(x, axis=None):
    """This enforces that the unbias estimate for variance is calculated"""
    # First make sure that the samples are in a numpy array
    return np.std(x, axis=1, ddof=1)


def unbiased_std(x, axis=None):
    """ A correction to numpy's standard deviation calculation.
    Calculate the unbiased estimation of standard deviation, which includes a correction
    factor for sample sizes < 100.
    """
    # First make sure that the samples are in a numpy array
    x = np.array(x)

    # What are the number of samples for this axis of interest
    if axis is None:
        n_samples = x.size
    else:
        n_samples = x.shape[axis]

    # If only 1 sample is given, just return numpy's normal standard deviation
    if n_samples is 1:
        return np.std(x, axis=axis)

    # Calculate the correction scalar
    c4 = c4_correction(n_samples)

    corrected_s = c4 * np.std(x, axis=axis, ddof=1)
    return corrected_s


def unbiased_sterror(x, axis=None):
    """ Unbiased error. """
    # First make sure that the samples are in a numpy array
    x = np.array(x)

    # What are the number of samples for this axis of interest
    if axis is None:
        n_samples = x.size
    else:
        n_samples = x.shape[axis]

    # Calculate biased standard deviation
    std = np.std(x, axis=axis)

    # Correct with c4 scalar
    c4 = c4_correction(n_samples)
    return std * np.sqrt(1 - c4)


# -----------------------------------------------------------------------
# Correction to sample statistics
# -----------------------------------------------------------------------

def corrected_std(var, n_samples=2):
    """Calculate the unbiased standard deviation from a biased standard deviation. """
    _var = np.array(var)
    _std = np.sqrt(_var)

    # If the sample size is 1, no correction applies
    if n_samples > 100:
        c4 = 1
    else:
        c4 = c4_correction(n_samples)
    return _std / c4


def corrected_sterror(var, n_samples=2):
    """Calculate an unbiased standard error from a BIASED standard deviation. """
    _var = np.array(var)

    # If sample size is 1, no correction applies
    if n_samples > 100:
        _std = np.sqrt(_var)
        sterr = _std / np.sqrt(n_samples)
    else:
        _std = np.sqrt(_var)
        sterr = _std / np.sqrt(n_samples)
        #_std = corrected_std(_var, n_samples=n_samples)
        #sterr = _std/np.sqrt(n_samples)

        #c4 = c4_correction(n_samples)
        #sterr = _std * np.sqrt(c4**(-2)-1)
    return sterr



from .nk import NKSimulation
from .base import random_mutation_set


class HouseOfCardsSimulation(NKSimulation):
    """Construct a 'House of Cards' fitness landscape.
    """

    def __init__(self, wildtype, mutations, k_range=(0, 1), *args, **kwargs):
        super(NKSimulation, self).__init__(wildtype, mutations, *args,
                                           **kwargs)
        # Set parameters
        self.set_order(len(self.binary[0]))
        self.set_random_values(k_range=k_range)
        self.build()
